use anyhow::Result;
use bytemuck;
use cgmath::{EuclideanSpace, Matrix2, Point2};
use std::{iter, rc::Rc};
use wgpu;
use winit::{self, dpi::PhysicalSize, window::Window};

use crate::Game;

use super::{
	globaldata::{GlobalData, GlobalDataContent},
	pipeline::PipelineBuilder,
	texturearray::TextureArray,
	vertexbuffer::{
		data::{SPRITE_INDICES, SPRITE_VERTICES},
		types::{SpriteInstance, StarfieldInstance, TexturedVertex},
		VertexBuffer,
	},
};

pub struct GPUState {
	device: wgpu::Device,
	config: wgpu::SurfaceConfiguration,
	surface: wgpu::Surface,
	queue: wgpu::Queue,

	pub window: Window,
	pub window_size: winit::dpi::PhysicalSize<u32>,
	window_aspect: f32,

	sprite_pipeline: wgpu::RenderPipeline,
	starfield_pipeline: wgpu::RenderPipeline,

	texture_array: TextureArray,
	global_data: GlobalData,
	vertex_buffers: VertexBuffers,
}

struct VertexBuffers {
	sprite: Rc<VertexBuffer>,
	starfield: Rc<VertexBuffer>,
}

impl GPUState {
	// We can draw at most this many sprites on the screen.
	// TODO: compile-time option
	pub const SPRITE_INSTANCE_LIMIT: u64 = 100;
	pub const STARFIELD_INSTANCE_LIMIT: u64 = 501 * 9;

	pub async fn new(window: Window) -> Result<Self> {
		let window_size = window.inner_size();
		let window_aspect = window_size.width as f32 / window_size.height as f32;

		let instance = wgpu::Instance::new(wgpu::InstanceDescriptor {
			backends: wgpu::Backends::all(),
			..Default::default()
		});

		let surface = unsafe { instance.create_surface(&window) }.unwrap();

		// Basic setup
		let device;
		let queue;
		let config;

		{
			let adapter = instance
				.request_adapter(&wgpu::RequestAdapterOptions {
					power_preference: wgpu::PowerPreference::default(),
					compatible_surface: Some(&surface),
					force_fallback_adapter: false,
				})
				.await
				.unwrap();

			(device, queue) = adapter
				.request_device(
					&wgpu::DeviceDescriptor {
						features: wgpu::Features::TEXTURE_BINDING_ARRAY | wgpu::Features::SAMPLED_TEXTURE_AND_STORAGE_BUFFER_ARRAY_NON_UNIFORM_INDEXING,
						// We may need limits if we compile for wasm
						limits: wgpu::Limits::default(),
						label: Some("gpu device"),
					},
					None,
				)
				.await
				.unwrap();

			// Assume sRGB
			let surface_caps = surface.get_capabilities(&adapter);
			let surface_format = surface_caps
				.formats
				.iter()
				.copied()
				.filter(|f| f.is_srgb())
				.filter(|f| f.has_stencil_aspect())
				.next()
				.unwrap_or(surface_caps.formats[0]);

			config = wgpu::SurfaceConfiguration {
				usage: wgpu::TextureUsages::RENDER_ATTACHMENT,
				format: surface_format,
				width: window_size.width,
				height: window_size.height,
				present_mode: surface_caps.present_modes[0],
				alpha_mode: surface_caps.alpha_modes[0],
				view_formats: vec![],
			};

			surface.configure(&device, &config);
		}

		let vertex_buffers = VertexBuffers {
			sprite: Rc::new(VertexBuffer::new::<TexturedVertex, SpriteInstance>(
				"sprite",
				&device,
				Some(SPRITE_VERTICES),
				Some(SPRITE_INDICES),
				Self::SPRITE_INSTANCE_LIMIT,
			)),

			starfield: Rc::new(VertexBuffer::new::<TexturedVertex, StarfieldInstance>(
				"starfield",
				&device,
				Some(SPRITE_VERTICES),
				Some(SPRITE_INDICES),
				Self::STARFIELD_INSTANCE_LIMIT,
			)),
		};

		// Load uniforms
		let global_data = GlobalData::new(&device);
		let texture_array = TextureArray::new(&device, &queue)?;

		// Make sure these match the indices in each shader
		let bind_group_layouts = &[
			&texture_array.bind_group_layout,
			&global_data.bind_group_layout,
		];

		// Create render pipelines
		let sprite_pipeline = PipelineBuilder::new("sprite", &device)
			.set_shader(include_str!(concat!(
				env!("CARGO_MANIFEST_DIR"),
				"/src/render/shaders/",
				"sprite.wgsl"
			)))
			.set_format(config.format)
			.set_triangle(true)
			.set_vertex_buffer(&vertex_buffers.sprite)
			.set_bind_group_layouts(bind_group_layouts)
			.build();

		let starfield_pipeline = PipelineBuilder::new("starfield", &device)
			.set_shader(include_str!(concat!(
				env!("CARGO_MANIFEST_DIR"),
				"/src/render/shaders/",
				"starfield.wgsl"
			)))
			.set_format(config.format)
			.set_triangle(true)
			.set_vertex_buffer(&vertex_buffers.starfield)
			.set_bind_group_layouts(bind_group_layouts)
			.build();

		return Ok(Self {
			device,
			config,
			surface,
			queue,

			window,
			window_size,
			window_aspect,

			sprite_pipeline,
			starfield_pipeline,

			texture_array,
			global_data,
			vertex_buffers,
		});
	}

	pub fn window(&self) -> &Window {
		&self.window
	}

	pub fn resize(&mut self, new_size: PhysicalSize<u32>) {
		if new_size.width > 0 && new_size.height > 0 {
			self.window_size = new_size;
			self.window_aspect = new_size.width as f32 / new_size.height as f32;
			self.config.width = new_size.width;
			self.config.height = new_size.height;
			self.surface.configure(&self.device, &self.config);
		}
	}

	pub fn update(&mut self) {}

	/// Make a SpriteInstance for each of the game's visible sprites.
	/// Will panic if SPRITE_INSTANCE_LIMIT is exceeded.
	///
	/// This is only called inside self.render()
	fn make_sprite_instances(&self, game: &Game) -> Vec<SpriteInstance> {
		let mut instances: Vec<SpriteInstance> = Vec::new();

		// Game coordinates (relative to camera) of ne and sw corners of screen.
		// Used to skip off-screen sprites.
		let clip_ne = Point2::from((-1.0, 1.0)) * game.camera.zoom;
		let clip_sw = Point2::from((1.0, -1.0)) * game.camera.zoom;

		for s in game.sprites() {
			let pos = s.post_parallax_position(game.camera) - game.camera.pos.to_vec();
			let texture = self.texture_array.get_texture(&s.name[..]);

			// Game dimensions of this sprite post-scale.
			//
			// We only need height / 2 to check if we're on the screen,
			// but we omit the division.
			// This gives us a small margin, and lets us re-use the value
			// without an extra multiply.
			let height = s.height * s.scale;
			let width = height * texture.aspect;

			// Don't draw (or compute matrices for)
			// sprites that are off the screen
			if pos.x < clip_ne.x - width
				|| pos.y > clip_ne.y + height
				|| pos.x > clip_sw.x + width
				|| pos.y < clip_sw.y - height
			{
				continue;
			}

			// Compute the values we need to draw
			// scale: combines texture scale and zoom scale.
			// screen_pos: position of this sprite in screen coordinates
			//
			// We can't use screen_pos to exclude off-screen sprites because
			// it can't account for height and width.

			instances.push(SpriteInstance {
				position: pos.into(),
				aspect: texture.aspect,
				rotation: Matrix2::from_angle(s.angle).into(),
				height,
				texture_index: texture.index,
			})
		}

		// Enforce sprite limit
		if instances.len() as u64 > Self::SPRITE_INSTANCE_LIMIT {
			// TODO: no panic, handle this better.
			unreachable!("Sprite limit exceeded!")
		}

		return instances;
	}

	/// Make a StarfieldInstance for each star that needs to be drawn.
	/// Will panic if STARFIELD_INSTANCE_LIMIT is exceeded.
	///
	/// This is only called inside self.render()
	fn make_starfield_instances(&self, game: &Game) -> Vec<StarfieldInstance> {
		let mut instances = Vec::new();
		for s in &game.system.starfield {
			// TODO: minimize operations here
			//let pos = Sprite::post_parallax(s.parallax, s.pos * 500.0, game.camera)
			//	- game.camera.pos.to_vec();
			instances.push(StarfieldInstance {
				position: s.pos.into(),
				parallax: s.parallax,
				height: 2.0,
			});
		}

		// Enforce starfield limit
		if instances.len() as u64 > Self::STARFIELD_INSTANCE_LIMIT {
			// TODO: no panic, handle this better.
			unreachable!("Starfield limit exceeded!")
		}

		return instances;
	}

	pub fn render(&mut self, game: &Game) -> Result<(), wgpu::SurfaceError> {
		let output = self.surface.get_current_texture()?;
		let view = output
			.texture
			.create_view(&wgpu::TextureViewDescriptor::default());

		let mut encoder = self
			.device
			.create_command_encoder(&wgpu::CommandEncoderDescriptor {
				label: Some("render encoder"),
			});

		let mut render_pass = encoder.begin_render_pass(&wgpu::RenderPassDescriptor {
			label: Some("render pass"),

			color_attachments: &[Some(wgpu::RenderPassColorAttachment {
				view: &view,
				resolve_target: None,
				ops: wgpu::Operations {
					load: wgpu::LoadOp::Clear(wgpu::Color {
						r: 0.0,
						g: 0.0,
						b: 0.0,
						a: 1.0,
					}),
					store: wgpu::StoreOp::Store,
				},
			})],
			depth_stencil_attachment: None,
			occlusion_query_set: None,
			timestamp_writes: None,
		});

		// Update global values
		self.queue.write_buffer(
			&self.global_data.buffer,
			0,
			bytemuck::cast_slice(&[GlobalDataContent {
				camera_position: game.camera.pos.into(),
				camera_zoom: game.camera.zoom,
				window_aspect: self.window_aspect,
				starfield_texture: 1,
				starfield_tile_size: 1000.0,
				padding: Default::default(),
			}]),
		);

		// Create sprite instances
		let sprite_instances = self.make_sprite_instances(game);
		self.queue.write_buffer(
			&self.vertex_buffers.sprite.instances,
			0,
			bytemuck::cast_slice(&sprite_instances),
		);

		// Create starfield instances
		let starfield_instances = self.make_starfield_instances(game);
		self.queue.write_buffer(
			&self.vertex_buffers.starfield.instances,
			0,
			bytemuck::cast_slice(&starfield_instances),
		);

		// These should match the indices in each shader,
		// and should each have a corresponding bind group layout.
		render_pass.set_bind_group(0, &self.texture_array.bind_group, &[]);
		render_pass.set_bind_group(1, &self.global_data.bind_group, &[]);

		// Starfield pipeline
		self.vertex_buffers.starfield.set_in_pass(&mut render_pass);
		render_pass.set_pipeline(&self.starfield_pipeline);

		render_pass.draw(0..1, 0..starfield_instances.len() as _);
		render_pass.draw_indexed(
			0..SPRITE_INDICES.len() as u32,
			0,
			0..starfield_instances.len() as _,
		);

		// Sprite pipeline
		self.vertex_buffers.sprite.set_in_pass(&mut render_pass);
		render_pass.set_pipeline(&self.sprite_pipeline);
		render_pass.draw_indexed(
			0..SPRITE_INDICES.len() as u32,
			0,
			0..sprite_instances.len() as _,
		);

		// begin_render_pass borrows encoder mutably, so we can't call finish()
		// without dropping this variable.
		drop(render_pass);

		self.queue.submit(iter::once(encoder.finish()));
		output.present();

		Ok(())
	}
}
