use anyhow::Result;
use serde::Deserialize;
use std::{collections::HashMap, fs::File, io::Read, num::NonZeroU32};
use wgpu::BindGroupLayout;

use super::rawtexture::RawTexture;

pub struct TextureArray {
	pub bind_group: wgpu::BindGroup,
	pub bind_group_layout: BindGroupLayout,
	textures: HashMap<String, Texture>,
	error_texture: Texture,
}

#[derive(Debug, Clone, Copy)]
pub struct Texture {
	pub index: u32,  // Index in texture array
	pub aspect: f32, // width / height
}

#[derive(Deserialize)]
struct Textures {
	texture: Vec<Tex>,
	error: Tex,
}

#[derive(Deserialize)]
struct Tex {
	name: String,
	path: String,
}

impl Tex {
	pub fn read(&self, device: &wgpu::Device, queue: &wgpu::Queue) -> Result<RawTexture> {
		let p = format!("{}/{}", TextureArray::INDEX_ROOT, self.path);
		let mut f = File::open(&p)?;
		let mut bytes = Vec::new();
		f.read_to_end(&mut bytes)?;
		RawTexture::from_bytes(&device, &queue, &bytes, &self.name)
	}
}

impl TextureArray {
	const INDEX_PATH: &'static str = "assets.ignore/index.toml";
	const INDEX_ROOT: &'static str = "assets.ignore";

	pub fn get_texture(&self, name: &str) -> Texture {
		match self.textures.get(name) {
			Some(x) => *x,
			None => self.error_texture,
		}
	}

	pub fn new(device: &wgpu::Device, queue: &wgpu::Queue) -> Result<Self> {
		// Load all textures

		let (error_texture, textures, texture_data) = {
			let mut texture_data: Vec<RawTexture> = Vec::new();
			let mut textures: HashMap<String, Texture> = HashMap::new();

			let mut texture_index_string = String::new();
			let _ = File::open(Self::INDEX_PATH)?.read_to_string(&mut texture_index_string);
			let texture_index: Textures = toml::from_str(&texture_index_string)?;

			// First texture is always error
			let raw = texture_index.error.read(device, queue)?;
			let error_texture = Texture {
				index: 0,
				aspect: raw.dimensions.0 as f32 / raw.dimensions.1 as f32,
			};
			texture_data.push(raw);

			// Load the rest
			let mut i = 1;
			for t in texture_index.texture {
				let raw = t.read(device, queue)?;
				let tx = Texture {
					index: i,
					aspect: raw.dimensions.0 as f32 / raw.dimensions.1 as f32,
				};
				textures.insert(t.name.clone(), tx);
				texture_data.push(raw);
				i += 1;
			}

			(error_texture, textures, texture_data)
		};

		let sampler = device.create_sampler(&wgpu::SamplerDescriptor {
			address_mode_u: wgpu::AddressMode::ClampToEdge,
			address_mode_v: wgpu::AddressMode::ClampToEdge,
			address_mode_w: wgpu::AddressMode::ClampToEdge,
			mag_filter: wgpu::FilterMode::Linear,
			min_filter: wgpu::FilterMode::Nearest,
			mipmap_filter: wgpu::FilterMode::Nearest,
			..Default::default()
		});

		let bind_group_layout = device.create_bind_group_layout(&wgpu::BindGroupLayoutDescriptor {
			label: Some("texture_bind_group_layout"),
			entries: &[
				// Texture data
				wgpu::BindGroupLayoutEntry {
					binding: 0,
					visibility: wgpu::ShaderStages::FRAGMENT,
					ty: wgpu::BindingType::Texture {
						multisampled: false,
						view_dimension: wgpu::TextureViewDimension::D2,
						sample_type: wgpu::TextureSampleType::Float { filterable: true },
					},
					count: NonZeroU32::new(texture_data.len() as u32),
				},
				// Texture sampler
				wgpu::BindGroupLayoutEntry {
					binding: 1,
					visibility: wgpu::ShaderStages::FRAGMENT,
					ty: wgpu::BindingType::Sampler(wgpu::SamplerBindingType::Filtering),
					count: NonZeroU32::new(texture_data.len() as u32),
				},
			],
		});

		let views: Vec<&wgpu::TextureView> = texture_data.iter().map(|x| &x.view).collect();
		let bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
			label: Some("texture_bind_group"),
			layout: &bind_group_layout,
			entries: &[
				wgpu::BindGroupEntry {
					binding: 0,
					// Array of all views
					resource: wgpu::BindingResource::TextureViewArray(&views),
				},
				wgpu::BindGroupEntry {
					binding: 1,
					resource: wgpu::BindingResource::SamplerArray(&[&sampler].repeat(views.len())),
				},
			],
		});

		return Ok(Self {
			bind_group,
			bind_group_layout,
			textures,
			error_texture,
		});
	}
}
