use std::mem;
use wgpu;

use super::BufferObject;

// Represents a textured vertex in WGSL
#[repr(C)]
#[derive(Copy, Clone, Debug, bytemuck::Pod, bytemuck::Zeroable)]
pub struct TexturedVertex {
	pub position: [f32; 3],
	pub texture_coords: [f32; 2],
}

impl BufferObject for TexturedVertex {
	fn layout() -> wgpu::VertexBufferLayout<'static> {
		wgpu::VertexBufferLayout {
			array_stride: Self::SIZE,
			step_mode: wgpu::VertexStepMode::Vertex,
			attributes: &[
				wgpu::VertexAttribute {
					offset: 0,
					shader_location: 0,
					format: wgpu::VertexFormat::Float32x3,
				},
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 3]>() as wgpu::BufferAddress,
					shader_location: 1,
					format: wgpu::VertexFormat::Float32x2,
				},
			],
		}
	}
}

// Represents a point in the starfield instance in WGSL
#[repr(C)]
#[derive(Copy, Clone, bytemuck::Pod, bytemuck::Zeroable)]
pub struct StarfieldInstance {
	/// Position in origin field tile.
	/// note that this is DIFFERENT from
	/// the way we provide sprite positions!
	pub position: [f32; 2],

	/// Parallax factor (same unit as usual)
	pub parallax: f32,

	/// Height of (unrotated) sprite in world units
	pub height: f32,
}

impl BufferObject for StarfieldInstance {
	fn layout() -> wgpu::VertexBufferLayout<'static> {
		wgpu::VertexBufferLayout {
			array_stride: Self::SIZE,
			step_mode: wgpu::VertexStepMode::Instance,
			attributes: &[
				// Position
				wgpu::VertexAttribute {
					offset: 0,
					shader_location: 2,
					format: wgpu::VertexFormat::Float32x2,
				},
				// Parallax
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 2]>() as wgpu::BufferAddress,
					shader_location: 3,
					format: wgpu::VertexFormat::Float32,
				},
				// Height
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 3]>() as wgpu::BufferAddress,
					shader_location: 4,
					format: wgpu::VertexFormat::Float32,
				},
			],
		}
	}
}

// Represents a sprite instance in WGSL
#[repr(C)]
#[derive(Copy, Clone, bytemuck::Pod, bytemuck::Zeroable)]
pub struct SpriteInstance {
	/// Rotation matrix for this sprite
	pub rotation: [[f32; 2]; 2],

	/// World position, relative to camera
	pub position: [f32; 2],

	/// Height of (unrotated) sprite in world units
	pub height: f32,

	// Sprite aspect ratio (width / height)
	pub aspect: f32,

	// What texture to use for this sprite
	pub texture_index: u32,
}

impl BufferObject for SpriteInstance {
	fn layout() -> wgpu::VertexBufferLayout<'static> {
		wgpu::VertexBufferLayout {
			array_stride: Self::SIZE,
			// Switch to a step mode of Vertex to Instance.
			// This means that our shaders will only change to use the next
			// instance when the shader starts processing a new instance
			step_mode: wgpu::VertexStepMode::Instance,
			attributes: &[
				// 2 arrays = 1 2x2 matrix
				wgpu::VertexAttribute {
					offset: 0,
					shader_location: 2,
					format: wgpu::VertexFormat::Float32x2,
				},
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 2]>() as wgpu::BufferAddress,
					shader_location: 3,
					format: wgpu::VertexFormat::Float32x2,
				},
				// Position
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 4]>() as wgpu::BufferAddress,
					shader_location: 4,
					format: wgpu::VertexFormat::Float32x2,
				},
				// Height
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 6]>() as wgpu::BufferAddress,
					shader_location: 5,
					format: wgpu::VertexFormat::Float32,
				},
				// Aspect
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 7]>() as wgpu::BufferAddress,
					shader_location: 6,
					format: wgpu::VertexFormat::Float32,
				},
				// Texture
				wgpu::VertexAttribute {
					offset: mem::size_of::<[f32; 8]>() as wgpu::BufferAddress,
					shader_location: 7,
					format: wgpu::VertexFormat::Uint32,
				},
			],
		}
	}
}
